# Copyright 2012 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# This script aims to configure an initial OpenStack environment with all the
# necessary configurations for tempest's run using nothing but OpenStack's
# native API.
# That includes, creating users, tenants, registering images (cirros),
# configuring neutron and so on.
#
# ASSUMPTION: this script is run by an admin user as it is meant to configure
# the OpenStack environment prior to actual use.

# Config
import ConfigParser
import os
import tarfile

from six.moves.urllib import request as urlreq

# Default client libs
import glanceclient as glance_client
import keystoneclient.v2_0.client as keystone_client

# Import OpenStack exceptions
import glanceclient.exc as glance_exception
import keystoneclient.exceptions as keystone_exception


TEMPEST_TEMP_DIR = os.getenv("TEMPEST_TEMP_DIR", "/tmp").rstrip('/')
TEMPEST_ROOT_DIR = os.getenv("TEMPEST_ROOT_DIR", os.getenv("HOME")).rstrip('/')

# Environment variables override defaults
TEMPEST_CONFIG_DIR = os.getenv("TEMPEST_CONFIG_DIR",
                               "%s%s" % (TEMPEST_ROOT_DIR, "/etc")).rstrip('/')
TEMPEST_CONFIG_FILE = os.getenv("TEMPEST_CONFIG_FILE",
                                "%s%s" % (TEMPEST_CONFIG_DIR, "/tempest.conf"))
TEMPEST_CONFIG_SAMPLE = os.getenv("TEMPEST_CONFIG_SAMPLE",
                                  "%s%s" % (TEMPEST_CONFIG_DIR,
                                            "/tempest.conf.sample"))
# Image references
IMAGE_DOWNLOAD_CHUNK_SIZE = 8 * 1024
IMAGE_UEC_SOURCE_URL = os.getenv("IMAGE_UEC_SOURCE_URL",
                                 "http://download.cirros-cloud.net/0.3.1/"
                                 "cirros-0.3.1-x86_64-uec.tar.gz")
TEMPEST_IMAGE_ID = os.getenv('IMAGE_ID')
TEMPEST_IMAGE_ID_ALT = os.getenv('IMAGE_ID_ALT')
IMAGE_STATUS_ACTIVE = 'active'


class ClientManager(object):
    """
    Manager that provides access to the official python clients for
    calling various OpenStack APIs.
    """
    def __init__(self):
        self.identity_client = None
        self.image_client = None
        self.network_client = None
        self.compute_client = None
        self.volume_client = None

    def get_identity_client(self, **kwargs):
        """
        Returns the openstack identity python client
        :param username: a string representing the username
        :param password: a string representing the user's password
        :param tenant_name: a string representing the tenant name of the user
        :param auth_url: a string representing the auth url of the identity
        :param insecure: True if we wish to disable ssl certificate validation,
        False otherwise
        :returns an instance of openstack identity python client
        """
        if not self.identity_client:
            self.identity_client = keystone_client.Client(**kwargs)

        return self.identity_client

    def get_image_client(self, version="1", *args, **kwargs):
        """
        This method returns OpenStack glance python client
        :param version: a string representing the version of the glance client
        to use.
        :param string endpoint: A user-supplied endpoint URL for the glance
                            service.
        :param string token: Token for authentication.
        :param integer timeout: Allows customization of the timeout for client
                                http requests. (optional)
        :return: a Client object representing the glance client
        """
        if not self.image_client:
            self.image_client = glance_client.Client(version, *args, **kwargs)

        return self.image_client


def get_tempest_config(path_to_config):
    """
    Gets the tempest configuration file as a ConfigParser object
    :param path_to_config: path to the config file
    :return: a ConfigParser object representing the tempest configuration file
    """
    # get the sample config file from the sample
    config = ConfigParser.ConfigParser()
    config.readfp(open(path_to_config))

    return config


def update_config_admin_credentials(config, config_section):
    """
    Updates the tempest config with the admin credentials
    :param config: a ConfigParser object representing the tempest config file
    :param config_section: the section name where the admin credentials are
    """
    # Check if credentials are present, default uses the config credentials
    OS_USERNAME = os.getenv('OS_USERNAME',
                            config.get(config_section, "admin_username"))
    OS_PASSWORD = os.getenv('OS_PASSWORD',
                            config.get(config_section, "admin_password"))
    OS_TENANT_NAME = os.getenv('OS_TENANT_NAME',
                               config.get(config_section, "admin_tenant_name"))
    OS_AUTH_URL = os.getenv('OS_AUTH_URL', config.get(config_section, "uri"))

    if not (OS_AUTH_URL and
            OS_USERNAME and
            OS_PASSWORD and
            OS_TENANT_NAME):
        raise Exception("Admin environment variables not found.")

    # TODO(tkammer): Add support for uri_v3
    config_identity_params = {'uri': OS_AUTH_URL,
                              'admin_username': OS_USERNAME,
                              'admin_password': OS_PASSWORD,
                              'admin_tenant_name': OS_TENANT_NAME}

    update_config_section_with_params(config,
                                      config_section,
                                      config_identity_params)


def update_config_section_with_params(config, config_section, params):
    """
    Updates a given config object with given params
    :param config: a ConfigParser object representing the tempest config file
    :param config_section: the section we would like to update
    :param params: the parameters we wish to update for that section
    """
    for option, value in params.items():
        config.set(config_section, option, value)


def get_identity_client_kwargs(config, config_section):
    """
    Get the required arguments for the identity python client
    :param config: a ConfigParser object representing the tempest config file
    :param config_section: the section name in the configuration where the
    arguments can be found
    :return: a dictionary representing the needed arguments for the identity
    client
    """
    username = config.get(config_section, 'admin_username')
    password = config.get(config_section, 'admin_password')
    tenant_name = config.get(config_section, 'admin_tenant_name')
    auth_url = config.get(config_section, 'uri')
    dscv = config.get(config_section, 'disable_ssl_certificate_validation')
    kwargs = {'username': username,
              'password': password,
              'tenant_name': tenant_name,
              'auth_url': auth_url,
              'insecure': dscv}

    return kwargs


def create_user_with_tenant(identity_client, username, password, tenant_name):
    """
    Creates a user using a given identity client
    :param identity_client: openstack identity python client
    :param username: a string representing the username
    :param password: a string representing the user's password
    :param tenant_name: a string representing the tenant name of the user
    """
    # Try to create the necessary tenant
    tenant_id = None
    try:
        tenant_description = "Tenant for Tempest %s user" % username
        tenant = identity_client.tenants.create(tenant_name,
                                                tenant_description)
        tenant_id = tenant.id
    except keystone_exception.Conflict:

        # if already exist, use existing tenant
        tenant_list = identity_client.tenants.list()
        for tenant in tenant_list:
            if tenant.name == tenant_name:
                tenant_id = tenant.id

    # Try to create the user
    try:
        email = "%s@test.com" % username
        identity_client.users.create(name=username,
                                     password=password,
                                     email=email,
                                     tenant_id=tenant_id)
    except keystone_exception.Conflict:

        # if already exist, use existing user
        pass


def create_users_and_tenants(identity_client,
                             config,
                             config_section):
    """
    Creates the two non admin users and tenants for tempest
    :param identity_client: openstack identity python client
    :param config: a ConfigParser object representing the tempest config file
    :param config_section: the section name of identity in the config
    """
    # Get the necessary params from the config file
    tenant_name = config.get(config_section, 'tenant_name')
    username = config.get(config_section, 'username')
    password = config.get(config_section, 'password')

    alt_tenant_name = config.get(config_section, 'alt_tenant_name')
    alt_username = config.get(config_section, 'alt_username')
    alt_password = config.get(config_section, 'alt_password')

    # Create the necessary users for the test runs
    create_user_with_tenant(identity_client, username, password, tenant_name)
    create_user_with_tenant(identity_client, alt_username, alt_password,
                            alt_tenant_name)


def get_image_client_kwargs(identity_client, config, config_section):
    """
    Get the required arguments for the image python client
    :param identity_client: openstack identity python client
    :param config: a ConfigParser object representing the tempest config file
    :param config_section: the section name of identity in the config
    :return: a dictionary representing the needed arguments for the image
    client
    """

    token = identity_client.auth_token
    endpoint = identity_client.\
        service_catalog.url_for(service_type='image', endpoint_type='publicURL'
                                )
    dscv = config.get(config_section, 'disable_ssl_certificate_validation')
    kwargs = {'endpoint': endpoint,
              'token': token,
              'insecure': dscv}

    return kwargs


def images_exist(image_client):
    """
    Checks whether the images ID's located in the environment variable are
    indeed registered
    :param image_client: the openstack python client representing the image
    client
    """
    exist = True
    if not TEMPEST_IMAGE_ID or not TEMPEST_IMAGE_ID_ALT:
        exist = False
    else:
        try:
            image_client.images.get(TEMPEST_IMAGE_ID)
            image_client.images.get(TEMPEST_IMAGE_ID_ALT)
        except glance_exception.HTTPNotFound:
            exist = False

    return exist


def download_and_register_uec_images(image_client, download_url,
                                     download_folder):
    """
    Downloads and registered the UEC AKI/AMI/ARI images
    :param image_client:
    :param download_url: the url of the uec tar file
    :param download_folder: the destination folder we wish to save the file to
    """
    basename = os.path.basename(download_url)
    path = os.path.join(download_folder, basename)

    request = urlreq.urlopen(download_url)

    # First, download the file
    with open(path, "wb") as fp:
        while True:
            chunk = request.read(IMAGE_DOWNLOAD_CHUNK_SIZE)
            if not chunk:
                break

            fp.write(chunk)

    # Then extract and register images
    tar = tarfile.open(path, "r")
    for name in tar.getnames():
        file_obj = tar.extractfile(name)
        format = "aki"

        if file_obj.name.endswith(".img"):
            format = "ami"

        if file_obj.name.endswith("initrd"):
            format = "ari"

        # Register images in image client
        image_client.images.create(name=file_obj.name, disk_format=format,
                                   container_format=format, data=file_obj,
                                   is_public="true")

    tar.close()


def create_images(image_client, config, config_section,
                  download_url=IMAGE_UEC_SOURCE_URL,
                  download_folder=TEMPEST_TEMP_DIR):
    """
    Creates images for tempest's use and registers the environment variables
    IMAGE_ID and IMAGE_ID_ALT with registered images
    :param image_client: OpenStack python image client
    :param config: a ConfigParser object representing the tempest config file
    :param config_section: the section name where the IMAGE ids are set
    :param download_url: the URL from which we should download the UEC tar
    :param download_folder: the place where we want to save the download file
    """
    if not images_exist(image_client):
        # Falls down to the default uec images
        download_and_register_uec_images(image_client, download_url,
                                         download_folder)
        image_ids = []
        for image in image_client.images.list():
            image_ids.append(image.id)

        os.environ["IMAGE_ID"] = image_ids[0]
        os.environ["IMAGE_ID_ALT"] = image_ids[1]

    params = {'image_ref': os.getenv("IMAGE_ID"),
              'image_ref_alt': os.getenv("IMAGE_ID_ALT")}

    update_config_section_with_params(config, config_section, params)


def main():
    """
    Main module to control the script
    """
    # Check if config file exists or fall to the default sample otherwise
    path_to_config = TEMPEST_CONFIG_SAMPLE

    if os.path.isfile(TEMPEST_CONFIG_FILE):
        path_to_config = TEMPEST_CONFIG_FILE

    config = get_tempest_config(path_to_config)
    update_config_admin_credentials(config, 'identity')

    client_manager = ClientManager()

    # Set the identity related info for tempest
    identity_client_kwargs = get_identity_client_kwargs(config,
                                                        'identity')
    identity_client = client_manager.get_identity_client(
        **identity_client_kwargs)

    # Create the necessary users and tenants for tempest run
    create_users_and_tenants(identity_client, config, 'identity')

    # Set the image related info for tempest
    image_client_kwargs = get_image_client_kwargs(identity_client,
                                                  config,
                                                  'identity')
    image_client = client_manager.get_image_client(**image_client_kwargs)

    # Create the necessary users and tenants for tempest run
    create_images(image_client, config, 'compute')

    # TODO(tkammer): add network implementation


if __name__ == "__main__":
    main()



# Copyright 2015 IBM Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import logging
import os
import requests
import subprocess
import tempfile

import httmock
from unittest import mock
import unittest

import refstack_client.list_parser as parser


class TestTestListParser(unittest.TestCase):

    test_path = os.path.dirname(os.path.realpath(__file__))
    tempest_dir = "some_dir/.tempest"

    def setUp(self):
        """Test case setup"""
        logging.disable(logging.CRITICAL)
        self.parser = parser.TestListParser(self.tempest_dir)

    def test_get_tempest_test_ids(self):
        """Test that the tempest test-list is correctly parsed."""
        test_list = ("tempest.test.one[gate]\n"
                     "tempest.test.two[gate,smoke]\n"
                     "tempest.test.three(scenario)\n"
                     "tempest.test.four[gate](another_scenario)\n"
                     "tempest.test.five")
        process_mock = mock.Mock(returncode=0)
        attrs = {'communicate.return_value': (test_list, None)}
        process_mock.configure_mock(**attrs)
        subprocess.Popen = mock.Mock(return_value=process_mock)
        output = self.parser._get_tempest_test_ids()

        subprocess.Popen.assert_called_with(
            ("%s/tools/with_venv.sh" % self.tempest_dir, "testr",
             "list-tests"),
            stdout=subprocess.PIPE,
            cwd=self.tempest_dir)
        expected_output = {"tempest.test.one": "[gate]",
                           "tempest.test.two": "[gate,smoke]",
                           "tempest.test.three(scenario)": "",
                           "tempest.test.four(another_scenario)": "[gate]",
                           "tempest.test.five": ""}
        self.assertEqual(expected_output, output)

    def test_get_tempest_test_ids_fail(self):
        """Test when the test listing subprocess returns a non-zero exit
        status.
        """
        process_mock = mock.Mock(returncode=1)
        attrs = {'communicate.return_value': (mock.ANY, None)}
        process_mock.configure_mock(**attrs)
        subprocess.Popen = mock.Mock(return_value=process_mock)
        with self.assertRaises(subprocess.CalledProcessError):
            self.parser._get_tempest_test_ids()

    def test_form_test_id_mappings(self):
        """Test the test ID to attribute dict builder function."""
        test_list = ["tempest.test.one[gate]",
                     "tempest.test.two[gate,smoke]",
                     "tempest.test.three(scenario)",
                     "tempest.test.four[gate](another_scenario)",
                     "tempest.test.five"]

        expected_output = {"tempest.test.one": "[gate]",
                           "tempest.test.two": "[gate,smoke]",
                           "tempest.test.three(scenario)": "",
                           "tempest.test.four(another_scenario)": "[gate]",
                           "tempest.test.five": ""}
        output = self.parser._form_test_id_mappings(test_list)
        self.assertEqual(expected_output, output)

    def test_get_base_test_ids_from_list_file(self):
        """test that we can get the base test IDs from a test list file."""
        list_file = self.test_path + "/test-list.txt"
        test_list = self.parser._get_base_test_ids_from_list_file(list_file)
        expected_list = ['tempest.api.test1',
                         'tempest.api.test2',
                         'tempest.api.test3(scenario)']
        self.assertEqual(expected_list, sorted(test_list))

    def test_get_base_test_ids_from_list_files_invalid_file(self):
        """Test that we get an exception when passing in a nonexistent file."""
        some_file = self.test_path + "/nonexistent.json"
        with self.assertRaises(Exception):
            self.parser._get_base_test_ids_from_list_file(some_file)

    def test_get_base_test_ids_from_list_file_url(self):
        """Test that we can parse the test cases from a test list URL."""
        list_file = self.test_path + "/test-list.txt"

        with open(list_file, 'rb') as f:
            content = f.read()

        @httmock.all_requests
        def request_mock(url, request):
            return {'status_code': 200,
                    'content': content}

        with httmock.HTTMock(request_mock):
            online_list = self.parser._get_base_test_ids_from_list_file(
                "http://127.0.0.1/test-list.txt")

        expected_list = ['tempest.api.test1',
                         'tempest.api.test2',
                         'tempest.api.test3(scenario)']
        self.assertEqual(expected_list, sorted(online_list))

    def test_get_base_test_ids_from_list_file_invalid_url(self):
        """Test a case of an invalid URL schema."""
        with self.assertRaises(requests.exceptions.RequestException):
            self.parser._get_base_test_ids_from_list_file("foo://sasas.com")

    def test_get_full_test_ids(self):
        """Test that full test IDs can be formed."""
        tempest_ids = {"tempest.test.one": "[gate]",
                       "tempest.test.two": "[gate,smoke]",
                       "tempest.test.three(scenario)": "",
                       "tempest.test.four(another_scenario)": "[gate]",
                       "tempest.test.five": ""}

        base_ids = ["tempest.test.one",
                    "tempest.test.four(another_scenario)",
                    "tempest.test.five"]

        output_list = self.parser._get_full_test_ids(tempest_ids, base_ids)
        expected_list = ["tempest.test.one[gate]",
                         "tempest.test.four[gate](another_scenario)",
                         "tempest.test.five"]
        self.assertEqual(expected_list, output_list)

    def test_get_full_test_ids_with_nonexistent_test(self):
        """Test when a test ID doesn't exist in the Tempest environment."""
        tempest_ids = {"tempest.test.one": "[gate]",
                       "tempest.test.two": "[gate,smoke]"}
        base_ids = ["tempest.test.one", "tempest.test.foo"]
        output_list = self.parser._get_full_test_ids(tempest_ids, base_ids)

        self.assertEqual(["tempest.test.one[gate]"], output_list)

    def test_write_normalized_test_list(self):
        """Test that a normalized test list is written to disk."""
        test_ids = ["tempest.test.one[gate]", "tempest.test.five"]
        test_file = self.parser._write_normalized_test_list(test_ids)

        # Check that the tempest IDs in the file match the expected test
        # ID list.
        with open(test_file, 'rb') as f:
            file_contents = f.read()
        testcase_list = list(filter(None,
                                    file_contents.decode('utf-8').split('\n')))

        self.assertEqual(test_ids, testcase_list)

    @mock.patch.object(parser.TestListParser, "get_normalized_test_list")
    def test_create_include_list(self, mock_get_normalized):
        """Test whether a test list is properly parsed to extract test names"""
        test_list = [
            "tempest.test.one[id-11111111-2222-3333-4444-555555555555,gate]",
            "tempest.test.two[comp,id-22222222-3333-4444-5555-666666666666]",
            "tempest.test.three[id-33333333-4444-5555-6666-777777777777](gate)"
        ]

        expected_list = "tempest.test.one\[\n"\
                        "tempest.test.two\[\n"\
                        "tempest.test.three\[\n" # noqa W605

        tmpfile = tempfile.mktemp()
        with open(tmpfile, 'w') as f:
            [f.write(item + "\n") for item in test_list]
        mock_get_normalized.return_value = tmpfile

        result = open(self.parser.create_include_list(tmpfile)).read()
        self.assertEqual(result, expected_list)



# Copyright 2014 IBM Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import os
import re
import subunit
import testtools
import unittest


class TempestSubunitTestResultPassOnly(testtools.TestResult):
    """Class to process subunit stream.

       This class maintains a list of test IDs that pass.
    """

    def __init__(self, stream, descriptions, verbosity):
        """Initialize with super class signature."""
        super(TempestSubunitTestResultPassOnly, self).__init__()
        self.results = []

    @staticmethod
    def get_test_uuid(test):
        attrs = None
        try:
            attrs = test.split('[')[1].split(']')[0].split(',')
        except IndexError:
            pass
        if not attrs:
            return
        for attr in attrs:
            if attr.startswith('id-'):
                return '-'.join(attr.split('-')[1:])

    def addSuccess(self, testcase):
        """Overwrite super class method for additional data processing."""
        super(TempestSubunitTestResultPassOnly, self).addSuccess(testcase)
        # Remove any [] from the test ID before appending it.
        # Will leave in any () for now as they are the only thing discerning
        # certain test cases.
        test_result = {'name': str(re.sub(r'\[.*\]', '', testcase.id()))}
        uuid = self.get_test_uuid(str(testcase.id()))
        if uuid:
            test_result['uuid'] = uuid
        self.results.append(test_result)

    def get_results(self):
        return self.results


class SubunitProcessor():
    """A class to replay subunit data from a stream."""

    def __init__(self, in_stream,
                 result_class=TempestSubunitTestResultPassOnly):
        self.in_stream = in_stream
        self.result_class = result_class

    def process_stream(self):
        """Read and process subunit data from a stream."""
        with open(self.in_stream, 'r') as fin:
            test = subunit.ProtocolTestCase(fin)
            runner = unittest.TextTestRunner(stream=open(os.devnull, 'w'),
                                             resultclass=self.result_class)

            # Run (replay) the test from subunit stream.
            test_result = runner.run(test)
            return test_result.get_results()



#
# Copyright (c) 2014 Piston Cloud Computing, Inc. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import hashlib
import json
import logging
import os
import tempfile

import httmock
from unittest import mock
from unittest.mock import MagicMock
import unittest

from refstack_client import refstack_client as rc


class TestRefstackClient(unittest.TestCase):

    test_path = os.path.dirname(os.path.realpath(__file__))
    conf_file_name = '%s/refstack-client.test.conf' % test_path

    def patch(self, name, **kwargs):
        """
        :param name: Name of class to be patched
        :param kwargs: directly passed to mock.patch
        :return: mock
        """
        patcher = mock.patch(name, **kwargs)
        thing = patcher.start()
        self.addCleanup(patcher.stop)
        return thing

    def mock_argv(self, command='test', **kwargs):
        """
        Build argv for test.
        :param conf_file_name: Configuration file name
        :param verbose: verbosity level
        :return: argv
        """
        argv = [command]
        if kwargs.get('verbose', None):
            argv.append(kwargs.get('verbose', None))
        if kwargs.get('silent', None):
            argv.append(kwargs.get('silent', None))
        argv.extend(['--url', 'http://127.0.0.1', '-y'])
        if kwargs.get('priv_key', None):
            argv.extend(('-i', kwargs.get('priv_key', None)))
        if command == 'test':
            argv.extend(
                ('-c', kwargs.get('conf_file_name', self.conf_file_name)))
            if kwargs.get('test_cases', None):
                argv.extend(('--', kwargs.get('test_cases', None)))
        return argv

    def mock_data(self):
        """
        Mock the Keystone client methods.
        """
        self.mock_identity_service_v2 = {'type': 'identity',
                                         'endpoints': [{'id': 'test-id'}]}
        self.mock_identity_service_v3 = {'type': 'identity',
                                         'id': 'test-id'}
        self.v2_config = {'auth_url': 'http://0.0.0.0:35357/v2.0/tokens',
                          'auth_version': 'v2',
                          'domain_name': 'Default',
                          'password': 'test',
                          'tenant_id': 'admin_project_id',
                          'project_id': 'admin_project_id',
                          'tenant_name': 'project_name',
                          'project_name': 'project_name',
                          'username': 'admin'}

    def setUp(self):
        """
        Test case setup
        """
        logging.disable(logging.CRITICAL)

    def test_verbose(self):
        """
        Test different verbosity levels.
        """
        args = rc.parse_cli_args(self.mock_argv())
        client = rc.RefstackClient(args)
        client.tempest_dir = self.test_path
        client._prep_test()
        self.assertEqual(client.logger.level, logging.INFO)

        args = rc.parse_cli_args(self.mock_argv(verbose='-v'))
        client = rc.RefstackClient(args)
        client.tempest_dir = self.test_path
        client._prep_test()
        self.assertEqual(client.logger.level, logging.DEBUG)

        args = rc.parse_cli_args(self.mock_argv(verbose='-vv'))
        client = rc.RefstackClient(args)
        client.tempest_dir = self.test_path
        client._prep_test()
        self.assertEqual(client.logger.level, logging.DEBUG)

        args = rc.parse_cli_args(self.mock_argv(silent='-s'))
        client = rc.RefstackClient(args)
        client.tempest_dir = self.test_path
        client._prep_test()
        self.assertEqual(client.logger.level, logging.WARNING)

    def test_get_next_stream_subunit_output_file(self):
        """
        Test getting the subunit file from an existing .testrepository
        directory that has a next-stream file.
        """
        args = rc.parse_cli_args(self.mock_argv())
        client = rc.RefstackClient(args)
        client.tempest_dir = self.test_path
        output_file = client._get_next_stream_subunit_output_file(
            self.test_path)

        # The next-stream file contains a "1".
        expected_file = expected_file = self.test_path + "/.testrepository/1"
        self.assertEqual(expected_file, output_file)

    def test_get_next_stream_subunit_output_file_nonexistent(self):
        """
        Test getting the subunit output file from a nonexistent
        .testrepository directory.
        """
        args = rc.parse_cli_args(self.mock_argv())
        client = rc.RefstackClient(args)
        output_file = client._get_next_stream_subunit_output_file(
            "/tempest/path")
        expected_file = "/tempest/path/.testrepository/0"
        self.assertEqual(expected_file, output_file)

    def test_get_cpid_account_file_not_found(self):
        """
        Test that the client will exit if an accounts file is specified,
        but does not exist.
        """
        args = rc.parse_cli_args(self.mock_argv())
        client = rc.RefstackClient(args)
        client.tempest_dir = self.test_path
        client._prep_test()

        client.conf.add_section('auth')
        client.conf.set('auth',
                        'test_accounts_file',
                        '%s/some-file.yaml' % self.test_path)

        self.mock_data()
        with self.assertRaises(SystemExit):
            client._get_keystone_config(client.conf)

    def test_get_keystone_config_account_file_empty(self):
        """
        Test that the client will exit if an accounts file exists,
        but is empty.
        """
        self.patch(
            'refstack_client.refstack_client.read_accounts_yaml',
            return_value=None)

        args = rc.parse_cli_args(self.mock_argv())
        client = rc.RefstackClient(args)
        client.tempest_dir = self.test_path
        client._prep_test()

        client.conf.add_section('auth')
        client.conf.set('auth',
                        'test_accounts_file',
                        '%s/some-file.yaml' % self.test_path)

        self.mock_data()
        with self.assertRaises(SystemExit):
            client._get_keystone_config(client.conf)

    def test_get_keystone_config_no_accounts_file(self):
        """
        Test that the client will exit if accounts file
        is not specified.
        """
        args = rc.parse_cli_args(self.mock_argv())
        client = rc.RefstackClient(args)
        client.tempest_dir = self.test_path
        client._prep_test()

        self.mock_data()
        with self.assertRaises(SystemExit):
            client._get_keystone_config(client.conf)

    def test_get_keystone_config(self):
        """
        Test that keystone configs properly parsed.
        """
        args = rc.parse_cli_args(self.mock_argv())
        client = rc.RefstackClient(args)
        client.tempest_dir = self.test_path
        client._prep_test()
        client.conf.add_section('auth')
        client.conf.set('auth',
                        'test_accounts_file',
                        '%s/test-accounts.yaml' % self.test_path)
        self.mock_data()
        accounts = [
            {
                'username': 'admin',
                'project_name': 'project_name',
                'project_id': 'admin_project_id',
                'password': 'test'
            }
        ]
        self.patch(
            'refstack_client.refstack_client.read_accounts_yaml',
            return_value=accounts)
        actual_result = client._get_keystone_config(client.conf)
        expected_result = self.v2_config
        self.assertEqual(expected_result, actual_result)

    def test_get_cpid_from_keystone_by_tenant_name_from_account_file(self):
        """
        Test getting a CPID from Keystone using an admin tenant name
        from an accounts file.
        """
        args = rc.parse_cli_args(self.mock_argv())
        client = rc.RefstackClient(args)
        client.tempest_dir = self.test_path
        client._prep_test()
        client.conf.add_section('auth')
        client.conf.set('auth',
                        'test_accounts_file',
                        '%s/test-accounts.yaml' % self.test_path)
        self.mock_data()
        actual_result = client._get_keystone_config(client.conf)
        expected_result = None
        self.assertEqual(expected_result, actual_result['tenant_id'])
        accounts = [
            {
                'username': 'admin',
                'tenant_id': 'tenant_id',
                'password': 'test'
            }
        ]
        self.patch(
            'refstack_client.refstack_client.read_accounts_yaml',
            return_value=accounts)
        actual_result = client._get_keystone_config(client.conf)
        self.assertEqual('tenant_id', actual_result['tenant_id'])

    def test_generate_keystone_data(self):
        """Test that correct data is generated."""
        args = rc.parse_cli_args(self.mock_argv())
        client = rc.RefstackClient(args)
        client.tempest_dir = self.test_path
        client._prep_test()
        client.conf.add_section('auth')
        client.conf.set('auth',
                        'test_accounts_file',
                        '%s/test-accounts.yaml' % self.test_path)
        self.mock_data()
        accounts = [
            {
                'username': 'admin',
                'tenant_id': 'admin_tenant_id',
                'password': 'test'
            }
        ]
        self.patch(
            'refstack_client.refstack_client.read_accounts_yaml',
            return_value=accounts)
        configs = client._get_keystone_config(client.conf)
        actual_results = client._generate_keystone_data(configs)
        expected_results = ('v2', 'http://0.0.0.0:35357/v2.0/tokens',
                            {'auth':
                                {'passwordCredentials':
                                    {
                                        'username': 'admin', 'password': 'test'
                                    },
                                 'tenantId': 'admin_tenant_id'}})
        self.assertEqual(expected_results, actual_results)

    def test_get_cpid_from_keystone_v3_varying_catalogs(self):
        """
        Test getting the CPID from keystone API v3 with varying catalogs.
        """
        args = rc.parse_cli_args(self.mock_argv())
        client = rc.RefstackClient(args)
        client.tempest_dir = self.test_path
        client._prep_test()
        client.conf.set('identity-feature-enabled', 'api_v3', 'true')
        client.conf.add_section('auth')
        client.conf.set('auth',
                        'test_accounts_file',
                        '%s/test-accounts.yaml' % self.test_path)
        self.mock_data()
        accounts = [
            {
                'tenant_name': 'tenant_name'
            }
        ]
        self.patch(
            'refstack_client.refstack_client.read_accounts_yaml',
            return_value=accounts)
        configs = client._get_keystone_config(client.conf)
        auth_version, auth_url, content = \
            client._generate_keystone_data(configs)
        client._generate_cpid_from_endpoint = MagicMock()

        # Test when the identity ID is None.
        ks3_ID_None = {'token': {'catalog':
                                 [{'type': 'identity', 'id': None}]}}

        @httmock.all_requests
        def keystone_api_v3_mock(url, request):
            return httmock.response(201, ks3_ID_None)
        with httmock.HTTMock(keystone_api_v3_mock):
            client._get_cpid_from_keystone(auth_version, auth_url, content)
            client._generate_cpid_from_endpoint.assert_called_with(auth_url)

        # Test when the catalog is empty.
        ks3_catalog_empty = {'token': {'catalog': []}}
        client._generate_cpid_from_endpoint = MagicMock()

        @httmock.all_requests
        def keystone_api_v3_mock2(url, request):
            return httmock.response(201, ks3_catalog_empty)
        with httmock.HTTMock(keystone_api_v3_mock2):
            client._get_cpid_from_keystone(auth_version, auth_url, content)
            client._generate_cpid_from_endpoint.assert_called_with(auth_url)

        # Test when there is no service catalog.
        ks3_no_catalog = {'token': {}}
        client._generate_cpid_from_endpoint = MagicMock()

        @httmock.all_requests
        def keystone_api_v3_mock3(url, request):
            return httmock.response(201, ks3_no_catalog)
        with httmock.HTTMock(keystone_api_v3_mock3):
            client._get_cpid_from_keystone(auth_version, auth_url, content)
            client._generate_cpid_from_endpoint.assert_called_with(auth_url)

        # Test when catalog has other non-identity services.
        ks3_other_services = {'token': {
                              'catalog': [{'type': 'compute',
                                           'id': 'test-id1'},
                                          {'type': 'identity',
                                           'id': 'test-id2'}]
                              }}
        client._generate_cpid_from_endpoint = MagicMock()

        @httmock.all_requests
        def keystone_api_v3_mock4(url, request):
            return httmock.response(201, ks3_other_services)
        with httmock.HTTMock(keystone_api_v3_mock4):
            cpid = client._get_cpid_from_keystone(auth_version,
                                                  auth_url,
                                                  content)
            self.assertFalse(client._generate_cpid_from_endpoint.called)
            self.assertEqual('test-id2', cpid)

    def test_get_cpid_from_keystone_failure_handled(self):
        """Test that get cpid from keystone API failure handled."""
        args = rc.parse_cli_args(self.mock_argv())
        client = rc.RefstackClient(args)
        client.tempest_dir = self.test_path
        client._prep_test()
        client.logger.warning = MagicMock()
        client._generate_cpid_from_endpoint = MagicMock()
        client.conf.add_section('auth')
        client.conf.set('auth',
                        'test_accounts_file',
                        '%s/test-accounts.yaml' % self.test_path)
        self.mock_data()
        accounts = [
            {
                'tenant_name': 'tenant_name',
                'tenant_id': 'admin_tenant_id',
                'password': 'test'
            }
        ]
        self.patch(
            'refstack_client.refstack_client.read_accounts_yaml',
            return_value=accounts)
        configs = client._get_keystone_config(client.conf)
        auth_version, url, content = client._generate_keystone_data(configs)

        @httmock.urlmatch(netloc=r'(.*\.)?127.0.0.1$', path='/v2/tokens')
        def keystone_api_mock(auth_version, url, request):
            return None
        with httmock.HTTMock(keystone_api_mock):
            client._get_cpid_from_keystone(auth_version, url, content)
            client._generate_cpid_from_endpoint.assert_called_with(url)

    def test_generate_cpid_from_endpoint(self):
        """
        Test that an endpoint's hostname is properly hashed.
        """
        args = rc.parse_cli_args(self.mock_argv())
        client = rc.RefstackClient(args)
        cpid = client._generate_cpid_from_endpoint('http://some.url:5000/v2')
        expected = hashlib.md5('some.url'.encode('utf-8')).hexdigest()
        self.assertEqual(expected, cpid)

        with self.assertRaises(ValueError):
            client._generate_cpid_from_endpoint('some.url:5000/v2')

    def test_form_result_content(self):
        """
        Test that the request content is formed into the expected format.
        """
        args = rc.parse_cli_args(self.mock_argv())
        client = rc.RefstackClient(args)
        content = client._form_result_content(1, 1, ['tempest.sample.test'])
        expected = {'cpid': 1,
                    'duration_seconds': 1,
                    'results': ['tempest.sample.test']}
        self.assertEqual(expected, content)

    def test_save_json_result(self):
        """
        Test that the results are properly written to a JSON file.
        """
        args = rc.parse_cli_args(self.mock_argv())
        client = rc.RefstackClient(args)
        results = {'cpid': 1,
                   'duration_seconds': 1,
                   'results': ['tempest.sample.test']}
        temp_file = tempfile.NamedTemporaryFile()
        client._save_json_results(results, temp_file.name)

        # Get the JSON that was written to the file and make sure it
        # matches the expected value.
        json_file = open(temp_file.name)
        json_data = json.load(json_file)
        json_file.close()
        self.assertEqual(results, json_data)

    def test_get_passed_tests(self):
        """
        Test that only passing tests are retrieved from a subunit file.
        """
        args = rc.parse_cli_args(self.mock_argv())
        client = rc.RefstackClient(args)
        subunit_file = self.test_path + "/.testrepository/0"
        results = client.get_passed_tests(subunit_file)
        expected = [
            {'name': 'tempest.passed.test'},
            {'name': 'tempest.tagged_passed.test',
             'uuid': '0146f675-ffbd-4208-b3a4-60eb628dbc5e'}
        ]
        self.assertEqual(expected, results)

    @mock.patch('six.moves.input')
    def test_user_query(self, mock_input):
        client = rc.RefstackClient(rc.parse_cli_args(self.mock_argv()))
        self.assertTrue(client._user_query('42?'))

        mock_input.return_value = 'n'
        cli_args = self.mock_argv()
        cli_args.remove('-y')
        client = rc.RefstackClient(rc.parse_cli_args(cli_args))
        self.assertFalse(client._user_query('42?'))
        mock_input.return_value = 'yes'
        self.assertTrue(client._user_query('42?'))

    def test_upload_prompt(self):
        """
        Test the _upload_prompt method.
        """
        client = rc.RefstackClient(rc.parse_cli_args(self.mock_argv()))

        # When user says yes.
        client._user_query = MagicMock(return_value=True)
        client.post_results = MagicMock()
        client._upload_prompt({'some': 'data'})
        client.post_results.assert_called_with(
            'http://127.0.0.1', {'some': 'data'}, sign_with=None
        )

        # When user says no.
        client._user_query = MagicMock(return_value=False)
        client.post_results = MagicMock()
        client._upload_prompt({'some': 'data'})
        self.assertFalse(client.post_results.called)

    def test_post_results(self):
        """
        Test the post_results method, ensuring a requests call is made.
        """
        args = rc.parse_cli_args(self.mock_argv())
        client = rc.RefstackClient(args)
        client.logger.info = MagicMock()
        content = {'duration_seconds': 0,
                   'cpid': 'test-id',
                   'results': [{'name': 'tempest.passed.test', 'uid': None}]}
        expected_response = json.dumps({'test_id': 42})

        @httmock.urlmatch(netloc=r'(.*\.)?127.0.0.1$', path='/v1/results/')
        def refstack_api_mock(url, request):
            return expected_response

        with httmock.HTTMock(refstack_api_mock):
            client.post_results("http://127.0.0.1", content)
            client.logger.info.assert_called_with(
                'http://127.0.0.1/v1/results/ Response: '
                '%s' % expected_response)

    def test_post_results_with_sign(self):
        """
        Test the post_results method, ensuring a requests call is made.
        """
        argv = self.mock_argv(command='upload', priv_key='rsa_key')
        argv.append('fake.json')
        args = rc.parse_cli_args(argv)
        client = rc.RefstackClient(args)
        client.logger.info = MagicMock()
        content = {'duration_seconds': 0,
                   'cpid': 'test-id',
                   'results': [{'name': 'tempest.passed.test'}]}
        expected_response = json.dumps({'test_id': 42})

        @httmock.urlmatch(netloc=r'(.*\.)?127.0.0.1$', path='/v1/results/')
        def refstack_api_mock(url, request):
            return expected_response

        with httmock.HTTMock(refstack_api_mock):
            rsapath = os.path.join(self.test_path, 'rsa_key')
            client.post_results("http://127.0.0.1", content, sign_with=rsapath)
            client.logger.info.assert_called_with(
                'http://127.0.0.1/v1/results/ Response: %s' %
                expected_response)

    def test_run_tempest(self):
        """
        Test that the test command will run the tempest script using the
        default configuration.
        """
        args = rc.parse_cli_args(
            self.mock_argv(verbose='-vv', test_cases='tempest.api.compute'))
        client = rc.RefstackClient(args)
        client.tempest_dir = self.test_path
        mock_popen = self.patch(
            'refstack_client.refstack_client.subprocess.Popen',
            return_value=MagicMock(returncode=0))
        self.patch("os.path.isfile", return_value=True)
        self.mock_data()
        client.get_passed_tests = MagicMock(return_value=[{'name': 'test'}])
        client.logger.info = MagicMock()
        client._save_json_results = MagicMock()
        client.post_results = MagicMock()
        client._get_keystone_config = MagicMock(
            return_value=self.v2_config)
        client.test()

        mock_popen.assert_called_with(
            ['%s/tools/with_venv.sh' % self.test_path, 'tempest', 'run',
             '--serial', '--regex', 'tempest.api.compute'],
            stderr=None
        )

        self.assertFalse(client.post_results.called)

    def test_run_tempest_upload(self):
        """
        Test that the test command will run the tempest script and call
        post_results when the --upload argument is passed in.
        """
        argv = self.mock_argv(verbose='-vv',
                              test_cases='tempest.api.compute')
        argv.insert(2, '--upload')
        args = rc.parse_cli_args(argv)
        client = rc.RefstackClient(args)
        client.tempest_dir = self.test_path
        mock_popen = self.patch(
            'refstack_client.refstack_client.subprocess.Popen',
            return_value=MagicMock(returncode=0))
        self.patch("os.path.isfile", return_value=True)
        self.mock_data()
        client.get_passed_tests = MagicMock(return_value=['test'])
        client.post_results = MagicMock()
        client._save_json_results = MagicMock()
        client._get_keystone_config = MagicMock(
            return_value=self.v2_config)
        client._get_cpid_from_keystone = MagicMock()
        client.test()
        mock_popen.assert_called_with(
            ['%s/tools/with_venv.sh' % self.test_path, 'tempest', 'run',
             '--serial', '--regex', 'tempest.api.compute'],
            stderr=None
        )

        self.assertTrue(client.post_results.called)

    def test_run_tempest_upload_with_sign(self):
        """
        Test that the test command will run the tempest script and call
        post_results when the --upload argument is passed in.
        """
        argv = self.mock_argv(verbose='-vv', priv_key='rsa_key',
                              test_cases='tempest.api.compute')
        argv.insert(2, '--upload')
        args = rc.parse_cli_args(argv)
        client = rc.RefstackClient(args)
        client.tempest_dir = self.test_path
        mock_popen = self.patch(
            'refstack_client.refstack_client.subprocess.Popen',
            return_value=MagicMock(returncode=0)
        )
        self.patch("os.path.isfile", return_value=True)
        self.mock_data()
        client.get_passed_tests = MagicMock(return_value=['test'])
        client.post_results = MagicMock()
        client._save_json_results = MagicMock()
        client._get_keystone_config = MagicMock(
            return_value=self.v2_config)
        client._get_cpid_from_keystone = MagicMock(
            return_value='test-id')
        client.test()
        mock_popen.assert_called_with(
            ['%s/tools/with_venv.sh' % self.test_path, 'tempest', 'run',
             '--serial', '--regex', 'tempest.api.compute'],
            stderr=None
        )

        self.assertTrue(client.post_results.called)
        client.post_results.assert_called_with(
            'http://127.0.0.1',
            {'duration_seconds': 0,
             'cpid': 'test-id',
             'results': ['test']},
            sign_with='rsa_key'
        )

    @mock.patch('refstack_client.list_parser.TestListParser.'
                'create_include_list')
    @mock.patch('refstack_client.list_parser.'
                'TestListParser.get_normalized_test_list')
    def test_run_tempest_with_test_list(self, mock_normalize,
                                        mock_include_list):
        """Test that the Tempest script runs with a test list file."""
        argv = self.mock_argv(verbose='-vv')
        argv.extend(['--test-list', 'test-list.txt'])
        args = rc.parse_cli_args(argv)
        client = rc.RefstackClient(args)
        client.tempest_dir = self.test_path
        mock_popen = self.patch(
            'refstack_client.refstack_client.subprocess.Popen',
            return_value=MagicMock(returncode=0))
        self.patch("os.path.isfile", return_value=True)
        self.patch("os.path.getsize", return_value=4096)
        self.mock_data()
        client.get_passed_tests = MagicMock(return_value=[{'name': 'test'}])
        client._save_json_results = MagicMock()
        client.post_results = MagicMock()
        mock_normalize.return_value = '/tmp/some-list'
        mock_include_list.return_value = '/tmp/some-list'
        client._get_keystone_config = MagicMock(
            return_value=self.v2_config)
        client.test()

        mock_include_list.assert_called_with('test-list.txt')
        # TODO(kopecmartin) rename the below argument when refstack-client
        # uses tempest which contains the following change in its code:
        # https://review.opendev.org/c/openstack/tempest/+/768583
        mock_popen.assert_called_with(
            ['%s/tools/with_venv.sh' % self.test_path, 'tempest', 'run',
             '--serial', '--whitelist_file', '/tmp/some-list'],
            stderr=None
        )

    def test_run_tempest_no_conf_file(self):
        """
        Test when a nonexistent configuration file is passed in.
        """
        args = rc.parse_cli_args(self.mock_argv(conf_file_name='ptn-khl'))
        client = rc.RefstackClient(args)
        self.assertRaises(SystemExit, client.test)

    def test_forbidden_conf_file(self):
        """
        Test when the user passes in a file that the user does not have
        read access to.
        """
        file = tempfile.NamedTemporaryFile()
        # Remove read access
        os.chmod(file.name, 0o220)
        args = rc.parse_cli_args(self.mock_argv(conf_file_name=file.name))
        client = rc.RefstackClient(args)
        self.assertRaises(SystemExit, client.test)

    def test_run_tempest_nonexisting_directory(self):
        """
        Test when the Tempest directory does not exist.
        """
        args = rc.parse_cli_args(self.mock_argv())
        client = rc.RefstackClient(args)
        client.tempest_dir = "/does/not/exist"
        self.assertRaises(SystemExit, client.test)

    def test_run_tempest_result_tag(self):
        """
        Check that the result JSON file is renamed with the result file tag
        when the --result-file-tag argument is passed in.
        """
        argv = self.mock_argv(verbose='-vv',
                              test_cases='tempest.api.compute')
        argv.insert(2, '--result-file-tag')
        argv.insert(3, 'my-test')
        args = rc.parse_cli_args(argv)
        client = rc.RefstackClient(args)
        client.tempest_dir = self.test_path
        mock_popen = self.patch(
            'refstack_client.refstack_client.subprocess.Popen',
            return_value=MagicMock(returncode=0))
        self.patch("os.path.isfile", return_value=True)
        self.mock_data()
        client.get_passed_tests = MagicMock(return_value=['test'])
        client._save_json_results = MagicMock()
        client._get_keystone_config = MagicMock(
            return_value=self.v2_config)
        client._get_cpid_from_keystone = MagicMock(
            return_value='test-id')
        client.test()

        mock_popen.assert_called_with(
            ['%s/tools/with_venv.sh' % self.test_path, 'tempest', 'run',
             '--serial', '--regex', 'tempest.api.compute'],
            stderr=None
        )
        # Since '1' is in the next-stream file, we expect the JSON output file
        # to be 'my-test-1.json'.
        expected_file = os.path.join(self.test_path, '.testrepository',
                                     'my-test-1.json')
        client._save_json_results.assert_called_with(mock.ANY, expected_file)

    def test_failed_run(self):
        """
        Test when the Tempest script returns a non-zero exit code.
        """
        self.patch('refstack_client.refstack_client.subprocess.Popen',
                   return_value=MagicMock(returncode=1))
        args = rc.parse_cli_args(self.mock_argv(verbose='-vv'))
        client = rc.RefstackClient(args)
        client.tempest_dir = self.test_path
        self.mock_data()
        client.logger.warning = MagicMock()
        client._get_keystone_config = MagicMock(
            return_value=self.v2_config)
        client._get_cpid_from_keystone = MagicMock()
        client.test()
        self.assertTrue(client.logger.warning.called)

    def test_upload(self):
        """
        Test that the upload command runs as expected.
        """
        upload_file_path = self.test_path + "/.testrepository/0.json"
        args = rc.parse_cli_args(
            self.mock_argv(command='upload', priv_key='rsa_key') +
            [upload_file_path])
        client = rc.RefstackClient(args)

        client.post_results = MagicMock()
        client.upload()
        expected_json = {
            'duration_seconds': 0,
            'cpid': 'test-id',
            'results': [
                {'name': 'tempest.passed.test'},
                {'name': 'tempest.tagged_passed.test',
                 'uuid': '0146f675-ffbd-4208-b3a4-60eb628dbc5e'}
            ]
        }
        client.post_results.assert_called_with('http://127.0.0.1',
                                               expected_json,
                                               sign_with='rsa_key')

    def test_subunit_upload(self):
        """
        Test that the subunit upload command runs as expected.
        """
        upload_file_path = self.test_path + "/.testrepository/0"
        args = rc.parse_cli_args(
            self.mock_argv(command='upload-subunit', priv_key='rsa_key') +
            ['--keystone-endpoint', 'http://0.0.0.0:5000/v2.0'] +
            [upload_file_path])
        client = rc.RefstackClient(args)
        client.post_results = MagicMock()
        client.upload_subunit()
        expected_json = {
            'duration_seconds': 0,
            'cpid': hashlib.md5('0.0.0.0'.encode('utf-8')).hexdigest(),
            'results': [
                {'name': 'tempest.passed.test'},
                {'name': 'tempest.tagged_passed.test',
                 'uuid': '0146f675-ffbd-4208-b3a4-60eb628dbc5e'}
            ]
        }
        client.post_results.assert_called_with('http://127.0.0.1',
                                               expected_json,
                                               sign_with='rsa_key')

    def test_upload_nonexisting_file(self):
        """
        Test when the file to be uploaded does not exist.
        """
        upload_file_path = self.test_path + "/.testrepository/foo.json"
        args = rc.parse_cli_args(['upload', upload_file_path,
                                  '--url', 'http://api.test.org'])
        client = rc.RefstackClient(args)
        self.assertRaises(SystemExit, client.upload)

    def test_yield_results(self):
        """
        Test the yield_results method, ensuring that results are retrieved.
        """
        args = rc.parse_cli_args(self.mock_argv(command='list'))
        client = rc.RefstackClient(args)
        expected_response = {
            "pagination": {
                "current_page": 1,
                "total_pages": 1
            },
            "results": [
                {
                    "cpid": "42",
                    "created_at": "2015-04-28 13:57:05",
                    "test_id": "1",
                    "url": "http://127.0.0.1:8000/output.html?test_id=1"
                },
                {
                    "cpid": "42",
                    "created_at": "2015-04-28 13:57:05",
                    "test_id": "2",
                    "url": "http://127.0.0.1:8000/output.html?test_id=2"
                }]}

        @httmock.urlmatch(netloc=r'(.*\.)?127.0.0.1$', path='/v1/results/')
        def refstack_api_mock(url, request):
            return json.dumps(expected_response)

        with httmock.HTTMock(refstack_api_mock):
            results = client.yield_results("http://127.0.0.1")
            self.assertEqual(expected_response['results'], next(results))
            # Since Python3.7 StopIteration exceptions are transformed into
            # RuntimeError (PEP 479):
            # https://docs.python.org/3/whatsnew/3.7.html
            self.assertRaises((StopIteration, RuntimeError), next, results)

    @mock.patch('six.moves.input', side_effect=KeyboardInterrupt)
    @mock.patch('sys.stdout', new_callable=MagicMock)
    def test_list(self, mock_stdout, mock_input):
        args = rc.parse_cli_args(self.mock_argv(command='list'))
        client = rc.RefstackClient(args)
        results = [[{"cpid": "42",
                     "created_at": "2015-04-28 13:57:05",
                     "test_id": "1",
                     "url": "http://127.0.0.1:8000/output.html?test_id=1"},
                    {"cpid": "42",
                     "created_at": "2015-04-28 13:57:05",
                     "test_id": "2",
                     "url": "http://127.0.0.1:8000/output.html?test_id=2"}]]
        mock_results = MagicMock()
        mock_results.__iter__.return_value = results
        client.yield_results = MagicMock(return_value=mock_results)
        client.list()
        self.assertTrue(mock_stdout.write.called)

    def test_sign_pubkey(self):
        """
        Test that the test command will run the tempest script and call
        post_results when the --upload argument is passed in.
        """
        args = rc.parse_cli_args(['sign',
                                  os.path.join(self.test_path, 'rsa_key')])
        client = rc.RefstackClient(args)
        pubkey, signature = client._sign_pubkey()
        self.assertTrue(pubkey.decode('utf8').startswith('ssh-rsa AAAA'))
        self.assertTrue(signature.decode('utf8').startswith('413cb954'))

    def test_set_env_params(self):
        """
        Test that the environment variables are correctly set.
        """
        args = rc.parse_cli_args(self.mock_argv())
        client = rc.RefstackClient(args)
        client.tempest_dir = self.test_path
        client._prep_test()
        conf_dir = os.path.abspath(os.path.dirname(self.conf_file_name))
        conf_file = os.path.basename(self.conf_file_name)
        self.assertEqual(os.environ.get('TEMPEST_CONFIG_DIR'), conf_dir)
        self.assertEqual(os.environ.get('TEMPEST_CONFIG'), conf_file)

    @mock.patch('refstack_client.list_parser.TestListParser.'
                'create_include_list')
    def test_run_tempest_with_empty_test_list(self, mock_include_list):
        """Test that refstack-client can handle an empty test list file."""
        argv = self.mock_argv(verbose='-vv')
        argv.extend(['--test-list', 'foo.txt'])
        args = rc.parse_cli_args(argv)
        client = rc.RefstackClient(args)
        self.mock_data()
        self.patch(
            'refstack_client.refstack_client.subprocess.Popen',
            return_value=MagicMock(returncode=0))
        client._get_keystone_config = MagicMock(return_value=self.v2_config)
        client.tempest_dir = self.test_path
        self.patch("os.path.isfile", return_value=True)
        empty_file = tempfile.NamedTemporaryFile()
        mock_include_list.return_value = empty_file.name
        self.assertRaises(SystemExit, client.test)

    def test_run_tempest_with_non_exist_test_list_file(self):
        """Test that refstack-client runs with a nonexistent test list file."""
        argv = self.mock_argv(verbose='-vv')
        argv.extend(['--test-list', 'foo.txt'])
        args = rc.parse_cli_args(argv)
        client = rc.RefstackClient(args)
        self.mock_data()
        self.patch(
            'refstack_client.list_parser.TestListParser._get_tempest_test_ids',
            return_value={'foo': ''})
        self.patch(
            'refstack_client.refstack_client.subprocess.Popen',
            return_value=MagicMock(returncode=0))
        client._get_keystone_config = MagicMock(return_value=self.v2_config)
        client.tempest_dir = self.test_path
        self.assertRaises(IOError, client.test)



# Copyright (c) 2015 IBM Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#


import atexit
import logging
import os
import re
import requests
import subprocess
import tempfile


class TestListParser(object):

    """This class is for normalizing test lists to match the tests in the
    current Tempest environment.
    """

    def __init__(self, tempest_dir, insecure=False):
        """
        Initialize the TestListParser.

        :param tempest_dir: Absolute path of the Tempest directory.
        :param insecure: Whether https requests, if any, should be insecure.
        """
        self.logger = logging.getLogger(__name__)
        self.tempest_dir = tempest_dir
        self.insecure = insecure

    def _get_tempest_test_ids(self):
        """This does a 'testr list-tests' or 'stestr list' according to
        Tempest version on the Tempest directory in order to get a list
        of full test IDs for the current Tempest environment. Test ID
        mappings are then formed for these tests.
        """
        if os.path.exists(os.path.join(self.tempest_dir, '.stestr.conf')):
            init_cmd = (os.path.join(self.tempest_dir, 'tools/with_venv.sh'),
                        'stestr', 'init')
            subprocess.Popen(init_cmd, stdout=subprocess.PIPE,
                             cwd=self.tempest_dir)
            cmd = (os.path.join(self.tempest_dir, 'tools/with_venv.sh'),
                   'stestr', 'list')
        else:
            cmd = (os.path.join(self.tempest_dir, 'tools/with_venv.sh'),
                   'testr', 'list-tests')
        process = subprocess.Popen(cmd, stdout=subprocess.PIPE,
                                   cwd=self.tempest_dir)
        (stdout, stderr) = process.communicate()

        if process.returncode != 0:
            self.logger.error(stdout)
            self.logger.error(stderr)
            raise subprocess.CalledProcessError(process.returncode,
                                                ' '.join(cmd))
        try:
            return self._form_test_id_mappings(stdout.split('\n'))
        except TypeError:
            return self._form_test_id_mappings(stdout.decode().split('\n'))

    def _form_test_id_mappings(self, test_list):
        """This takes in a list of full test IDs and forms a dict containing
        base test IDs mapped to their attributes. A full test ID also contains
        test attributes such as '[gate,smoke]'
        Ex:
            'tempest.api.test1': '[gate]'
            'tempest.api.test2': ''
            'tempest.api.test3(some_scenario)': '[smoke,gate]'

        :param test_list: List of full test IDs
        """
        test_mappings = {}
        for testcase in test_list:
            if testcase != "":
                # Search for any strings like '[smoke, gate]' in the test ID.
                match = re.search(r'(\[.*\])', testcase)

                if match:
                    testcase = re.sub(r'\[.*\]', '', testcase)
                    test_mappings[testcase] = match.group(1)
                else:
                    test_mappings[testcase] = ""
        return test_mappings

    def _get_base_test_ids_from_list_file(self, list_location):
        """This takes in a test list file and finds all the base test IDs
        for the tests listed.
        Ex:
            'tempest.test1[gate,id-2]' -> 'tempest.test1'
            'tempest.test2[gate,id-3](scenario)' -> 'tempest.test2(scenario)'

        :param list_location: file path or URL location of list file
        """
        try:
            response = requests.get(list_location,
                                    verify=not self.insecure)
            testcase_list = response.text.split('\n')
            test_mappings = self._form_test_id_mappings(testcase_list)
        # If the location isn't a valid URL, we assume it is a file path.
        except requests.exceptions.MissingSchema:
            try:
                with open(list_location) as data_file:
                    testcase_list = [line.rstrip('\n') for line in data_file]
                test_mappings = self._form_test_id_mappings(testcase_list)
            except Exception:
                self.logger.error("Error reading the passed in test list " +
                                  "file.")
                raise
        except Exception:
            self.logger.error("Error reading the passed in test list file.")
            raise

        return list(test_mappings.keys())

    def _get_full_test_ids(self, tempest_ids, base_ids):
        """This will remake the test ID list with the full IDs of the current
        Tempest environment. The Tempest test ID dict should have the correct
        mappings.

        :param tempest_ids: dict containing test ID mappings
        :param base_ids: list containing base test IDs
        """
        test_list = []
        for test_id in base_ids:
            try:
                attr = tempest_ids[test_id]
                # If the test has a scenario in the test ID, but also has some
                # additional attributes, the attributes need to go before the
                # scenario.
                if '(' in test_id and attr:
                    components = test_id.split('(', 1)
                    test_portion = components[0]
                    scenario = "(" + components[1]
                    test_list.append(test_portion + attr + scenario)
                else:
                    test_list.append(test_id + attr)
            except KeyError:
                self.logger.debug("Test %s not found in Tempest list." %
                                  test_id)
        self.logger.debug("Number of tests: " + str(len(test_list)))
        return test_list

    def _write_normalized_test_list(self, test_ids):
        """Create a temporary file to pass into testr containing a list of test
        IDs that should be tested.

        :param test_ids: list of full test IDs
        """
        temp = tempfile.NamedTemporaryFile(delete=False)
        for test_id in test_ids:
            temp.write(("%s\n" % test_id).encode('utf-8'))
        temp.flush()

        # Register the created file for cleanup.
        atexit.register(self._remove_test_list_file, temp.name)
        return temp.name

    def _remove_test_list_file(self, file_path):
        """Delete the given file.

        :param file_path: string containing the location of the file
        """
        if os.path.isfile(file_path):
            os.remove(file_path)

    def get_normalized_test_list(self, list_location):
        """This will take in the user's test list and will normalize it
        so that the test cases in the list map to actual full test IDS in
        the Tempest environment.

        :param list_location: file path or URL of the test list
        """
        tempest_test_ids = self._get_tempest_test_ids()
        if not tempest_test_ids:
            return None
        base_test_ids = self._get_base_test_ids_from_list_file(list_location)
        full_capability_test_ids = self._get_full_test_ids(tempest_test_ids,
                                                           base_test_ids)
        list_file = self._write_normalized_test_list(full_capability_test_ids)
        return list_file

    def create_include_list(self, list_location):
        """This takes in a test list file, get normalized, and get list of
        include regexes using full qualified test names (one per line).
        Ex:
          'tempest.test1[id-2,gate]' -> tempest.test1\[ # noqa: W605
          'tempest.test2[id-3,smoke](scenario)' -> tempest.test2\[ # noqa: W605
          'tempest.test3[compute,id-4]' -> tempest.test3\[ # noqa: W605

        :param list_location: file path or URL location of list file
        """
        normalized_list = open(self.get_normalized_test_list(list_location),
                               'r').read()
        # Keep the names
        tests_list = [re.sub(r"\[", r"\[", test)
                      for test in re.findall(r".*\[", normalized_list)]

        return self._write_normalized_test_list(tests_list)



#!/usr/bin/env python
#
# Copyright (c) 2014 Piston Cloud Computing, Inc. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#


"""
Run Tempest and upload results to RefStack.

This module runs the Tempest test suite on an OpenStack environment given a
Tempest configuration file.

"""

import argparse
import binascii
import hashlib
import itertools
import json
import logging
import os
import subprocess
import time

from cryptography.exceptions import UnsupportedAlgorithm
from cryptography.hazmat import backends
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import padding

from config_tempest import main
from config_tempest import constants as C
from keystoneauth1 import exceptions as KE
from openstack import exceptions as OSE

import requests
import requests.exceptions
from six import moves
from six.moves.urllib import parse
from refstack_client.subunit_processor import SubunitProcessor
from refstack_client.list_parser import TestListParser
import yaml


def get_input():
    """
    Wrapper for raw_input. Necessary for testing.
    """
    return moves.input().lower()  # pragma: no cover


def read_accounts_yaml(path):
    """Reads a set of accounts from the specified file"""
    with open(path, 'r') as yaml_file:
        accounts = yaml.safe_load(yaml_file)
    return accounts


class RefstackClient:
    log_format = "%(asctime)s %(name)s:%(lineno)d %(levelname)s %(message)s"

    def __init__(self, args):
        '''Prepare a tempest test against a cloud.'''
        self.logger = logging.getLogger("refstack_client")
        self.console_log_handle = logging.StreamHandler()
        self.console_log_handle.setFormatter(
            logging.Formatter(self.log_format))
        self.logger.addHandler(self.console_log_handle)

        self.args = args
        self.current_dir = os.path.dirname(os.path.realpath(__file__))
        self.refstack_dir = os.path.dirname(self.current_dir)
        self.tempest_dir = os.path.join(self.refstack_dir, '.tempest')

        # set default log level to INFO.
        if self.args.silent:
            self.logger.setLevel(logging.WARNING)
        elif self.args.verbose:
            self.logger.setLevel(logging.DEBUG)
        else:
            self.logger.setLevel(logging.INFO)

    def _prep_test(self):
        '''Prepare a tempest test against a cloud.'''

        # Check that the config file exists.
        if not os.path.isfile(self.args.conf_file):
            self.logger.error("Conf file not valid: %s" % self.args.conf_file)
            exit(1)

        if not os.access(self.args.conf_file, os.R_OK):
            self.logger.error("You do not have read access to: %s"
                              % self.args.conf_file)
            exit(1)

        # Initialize environment variables with config file info
        os.environ["TEMPEST_CONFIG_DIR"] = os.path.abspath(
            os.path.dirname(self.args.conf_file))
        os.environ["TEMPEST_CONFIG"] = os.path.basename(self.args.conf_file)

        # Check that the Tempest directory is an existing directory.
        if not os.path.isdir(self.tempest_dir):
            self.logger.error("Tempest directory given is not a directory or "
                              "does not exist: %s" % self.tempest_dir)
            exit(1)

        self.conf_file = self.args.conf_file
        # Note: SafeConfigParser deprecated on Python 3.2
        # Use ConfigParser directly
        self.conf = moves.configparser.ConfigParser()
        self.conf.read(self.args.conf_file)

    def _prep_upload(self):
        '''Prepare an upload to the RefStack_api'''
        if not os.path.isfile(self.args.file):
            self.logger.error("File not valid: %s" % self.args.file)
            exit(1)

        self.upload_file = self.args.file

    def _get_next_stream_subunit_output_file(self, tempest_dir):
        '''This method reads from the next-stream file in the .testrepository
           or .stestr directory according to Tempest version of the given
           Tempest path. The integer here is the name of the file where subunit
           output will be saved to. It also checks if the repository is
           initialized and if not, initializes it'''
        if os.path.exists(os.path.join(tempest_dir, '.stestr.conf')):
            sub_dir = '.stestr'
            cmd = 'stestr'
        else:
            sub_dir = '.testrepository'
            cmd = 'testr'
        try:
            if not os.path.exists(os.path.join(tempest_dir, sub_dir)):
                self.logger.debug('No repository found, creating one.')
                os.chdir(tempest_dir)
                process = subprocess.Popen([cmd, 'init'])
                process.communicate()
                os.chdir(self.refstack_dir)

            subunit_file = open(os.path.join(
                                tempest_dir, sub_dir,
                                'next-stream'), 'r').read().rstrip()
        except (IOError, OSError):
            self.logger.debug('The ' + sub_dir + '/next-stream file was not '
                              'found. Assuming subunit results will be stored '
                              'in file 0.')

            # First test stream is saved into $sub_dir/0
            subunit_file = "0"

        return os.path.join(tempest_dir, sub_dir, subunit_file)

    def _get_keystone_config(self, conf_file):
        '''This will get and return the keystone configs from config file.'''
        try:
            # Prefer Keystone V3 API if it is enabled
            auth_version = 'v3'
            if conf_file.has_option('identity', 'auth_version'):
                auth_version = conf_file.get('identity', 'auth_version')

            if auth_version == 'v2':
                auth_url = '%s/tokens' % (conf_file.get('identity', 'uri')
                                          .rstrip('/'))
            elif auth_version == 'v3':
                auth_url = '%s/auth/tokens' % (conf_file.get('identity',
                                               'uri_v3').rstrip('/'))

            domain_name = 'Default'
            if conf_file.has_option('identity', 'domain_name'):
                domain_name = conf_file.get('identity', 'domain_name')
            if conf_file.has_option('auth', 'test_accounts_file'):
                account_file = os.path.expanduser(
                    conf_file.get('auth', 'test_accounts_file'))
                if not os.path.isfile(account_file):
                    self.logger.error(
                        'Accounts file not found: %s' % account_file)
                    exit(1)

                accounts = read_accounts_yaml(account_file)
                if not accounts:
                    self.logger.error('Accounts file %s found, '
                                      'but was empty.' % account_file)
                    exit(1)
                account = accounts[0]
                username = account.get('username')
                password = account.get('password')
                project_id = (account.get('tenant_id') or
                              account.get('project_id'))
                project_name = (account.get('tenant_name') or
                                account.get('project_name'))
                return {'auth_version': auth_version,
                        'auth_url': auth_url,
                        'domain_name': domain_name,
                        'username': username, 'password': password,
                        'tenant_id': project_id, 'tenant_name': project_name,
                        'project_id': project_id, 'project_name': project_name
                        }
            elif conf_file.has_option('identity', 'username'):
                self.logger.error('Using identity section in tempest config '
                                  'file to specify user credentials is no '
                                  'longer supported. User credentials should '
                                  'be defined in the accounts file as '
                                  'described in the Tempest configuration '
                                  'guide (https://docs.openstack.org/tempest/'
                                  'latest/configuration.html).')
                exit(1)
            else:
                self.logger.error('User credentials cannot be found. '
                                  'User credentials should be defined in the '
                                  'accounts file as described in the Tempest '
                                  'configuration guide (http://docs.openstack.'
                                  'org/tempest/latest/configuration.html).')
                exit(1)
        except moves.configparser.Error as e:
            # Most likely a missing section or option in the config file.
            self.logger.error("Invalid Config File: %s" % e)
            exit(1)

    def _generate_keystone_data(self, auth_config):
        '''This will generate data for http post to keystone
        API from auth_config.'''
        auth_version = auth_config['auth_version']
        auth_url = auth_config['auth_url']
        if auth_version == 'v2':
            password_credential = {'username': auth_config['username'],
                                   'password': auth_config['password']}
            if auth_config['tenant_id']:
                data = {
                    'auth': {
                        'tenantId': auth_config['tenant_id'],
                        'passwordCredentials': password_credential
                    }
                }
            else:
                data = {
                    'auth': {
                        'tenantName': auth_config['tenant_name'],
                        'passwordCredentials': password_credential
                    }
                }
            return auth_version, auth_url, data
        elif auth_version == 'v3':
            identity = {
                'methods': ['password'],
                'password': {
                    'user': {
                        'name': auth_config['username'],
                        'domain': {'name': auth_config['domain_name']},
                        'password': auth_config['password']
                    }}}

            data = {
                'auth': {
                    'identity': identity,
                    'scope': {
                        'project': {
                            'name': auth_config['tenant_name'],
                            'domain': {'name': auth_config['domain_name']}
                        }
                    }
                }
            }
            return auth_version, auth_url, data

    def _get_cpid_from_keystone(self, auth_version, auth_url, content):
        '''This will get the Keystone service ID which is used as the CPID.'''
        try:
            headers = {'content-type': 'application/json'}
            response = requests.post(auth_url,
                                     data=json.dumps(content),
                                     headers=headers,
                                     verify=not self.args.insecure)
            rsp = response.json()
            if response.status_code in (200, 203):
                # keystone API v2 response.
                access = rsp['access']
                for service in access['serviceCatalog']:
                    if service['type'] == 'identity':
                        if service['endpoints'][0]['id']:
                            return service['endpoints'][0]['id']
                # Raise a key error if 'identity' was not found so that it
                # can be caught and have an appropriate error displayed.
                raise KeyError
            elif response.status_code == 201:
                # keystone API v3 response.
                token = rsp['token']
                for service in token['catalog']:
                    if service['type'] == 'identity' and service['id']:
                        return service['id']
                # Raise a key error if 'identity' was not found.
                # It will be caught below as well.
                raise KeyError
            else:
                message = ('Invalid request with error '
                           'code: %s. Error message: %s'
                           '' % (rsp['error']['code'],
                                 rsp['error']['message']))
                raise requests.exceptions.HTTPError(message)
            # If a Key or Index Error was raised, one of the expected keys or
            # indices for retrieving the identity service ID was not found.
        except (KeyError, IndexError):
            self.logger.warning('Unable to retrieve CPID from Keystone %s '
                                'catalog. The catalog or the identity '
                                'service endpoint was not '
                                'found.' % auth_version)
        except Exception as e:
            self.logger.warning('Problems retrieving CPID from Keystone '
                                'using %s endpoint (%s) with error (%s)'
                                % (auth_version, auth_url, e))
        return self._generate_cpid_from_endpoint(auth_url)

    def _generate_cpid_from_endpoint(self, endpoint):
        '''This method will md5 hash the hostname of a Keystone endpoint to
           generate a CPID.'''
        self.logger.info('Creating hash from endpoint to use as CPID.')
        url_parts = parse.urlparse(endpoint)
        if url_parts.scheme not in ('http', 'https'):
            raise ValueError('Invalid Keystone endpoint format. Make sure '
                             'the endpoint (%s) includes the URL scheme '
                             '(i.e. http/https).' % endpoint)
        return hashlib.md5(url_parts.hostname.encode('utf-8')).hexdigest()

    def _form_result_content(self, cpid, duration, results):
        '''This method will create the content for the request. The spec at
           'https://opendev.org/openinfra/refstack/src/branch/master/specs/prior'
           '/implemented/api-v1.md'.
           defines the format expected by the API.'''
        content = {}
        content['cpid'] = cpid
        content['duration_seconds'] = duration
        content['results'] = results
        return content

    def _save_json_results(self, results, path):
        '''Save the output results from the Tempest run as a JSON file'''
        file = open(path, "w+")
        file.write(json.dumps(results, indent=4, separators=(',', ': ')))
        file.close()

    def _user_query(self, q):
        """Ask user a query. Return true if user agreed (yes/y)"""
        if self.args.quiet:
            return True
        try:
            inp = moves.input(q + ' (yes/y): ')
        except KeyboardInterrupt:
            return
        else:
            return inp.lower() in ('yes', 'y')

    def _upload_prompt(self, upload_content):
        if self._user_query('Test results will be uploaded to %s. '
                            'Ok?' % self.args.url):
            self.post_results(self.args.url, upload_content,
                              sign_with=self.args.priv_key)

    def get_passed_tests(self, result_file):
        '''Get a list of tests IDs that passed Tempest from a subunit file.'''
        subunit_processor = SubunitProcessor(result_file)
        results = subunit_processor.process_stream()
        return results

    def post_results(self, url, content, sign_with=None):
        '''Post the combined results back to the refstack server.'''
        endpoint = '%s/v1/results/' % url
        headers = {'Content-type': 'application/json'}
        data = json.dumps(content)
        self.logger.debug('API request content: %s ' % content)
        if sign_with:
            with open(sign_with) as private_key_file:
                try:
                    private_key = serialization.load_pem_private_key(
                        private_key_file.read().encode('utf-8'),
                        password=None,
                        backend=backends.default_backend())
                except (IOError, UnsupportedAlgorithm, ValueError) as e:
                    self.logger.info('Error during upload key pair %s' %
                                     private_key_file)
                    self.logger.exception(e)
                    return
            signature = private_key.sign(
                data.encode('utf-8'),
                padding.PKCS1v15(),
                hashes.SHA256()
            )
            pubkey = private_key.public_key().public_bytes(
                serialization.Encoding.OpenSSH,
                serialization.PublicFormat.OpenSSH)

            headers['X-Signature'] = binascii.b2a_hex(signature)
            headers['X-Public-Key'] = pubkey
        try:
            response = requests.post(endpoint,
                                     data=data,
                                     headers=headers,
                                     verify=not self.args.insecure)
            self.logger.info(endpoint + " Response: " + str(response.text))
        except Exception as e:
            self.logger.info('Failed to post %s - %s ' % (endpoint, e))
            self.logger.exception(e)
            return

        if response.status_code == 201:
            resp = response.json()
            print('Test results uploaded!\nURL: %s' % resp.get('url', ''))

    def generate_tempest_config(self):
        '''Generate tempest.conf for a deployed OpenStack Cloud.'''
        start_time = time.time()

        # Write tempest.conf in refstack_client/etc folder
        if not self.args.out:
            config_path = os.path.join(self.refstack_dir, 'etc/tempest.conf')
        else:
            config_path = self.args.out
        self.logger.info("Generating in %s" % config_path)

        # Generate Tempest configuration
        try:
            cloud_creds = main.get_cloud_creds(self.args)
        except KE.MissingRequiredOptions as e:
            self.logger.error("Credentials are not sourced - %s" % e)
        except OSE.ConfigException:
            self.logger.error("Named cloud %s was not found"
                              % self.args.os_cloud)

        # tempestconf arguments
        kwargs = {'non_admin': True,
                  'test_accounts': self.args.test_accounts,
                  'image_path': self.args.image,
                  'network_id': self.args.network_id,
                  'out': config_path,
                  'cloud_creds': cloud_creds}

        # Look for extra overrides to be replaced in tempest.conf
        # (TODO:chkumar246) volume-feature-enabled.api_v2=True is deprecated
        # in ROCKY release, but it is required for interop tests and it is out
        # of scope of python-tempestconf, adding it hardcoded here as a extra
        # overrides.
        cinder_overrides = "volume-feature-enabled.api_v2=True"
        overrides_format = cinder_overrides.replace('=', ',').split(',')
        overrides = []
        if self.args.overrides:
            overrides = self.args.overrides.replace('=', ',').split(',')
            if cinder_overrides not in self.args.overrides:
                overrides = overrides + overrides_format
        else:
            overrides = overrides_format
        kwargs.update({'overrides': main.parse_overrides(overrides)})

        # Generate accounts.yaml if accounts.file is not given
        if not self.args.test_accounts:
            account_file = os.path.join(self.refstack_dir, 'etc/accounts.yaml')
            kwargs.update({'create_accounts_file': account_file})
            self.logger.info('Account file will be generated at %s.'
                             % account_file)

        # Generate tempest.conf
        main.config_tempest(**kwargs)

        if os.path.isfile(config_path):
            end_time = time.time()
            elapsed = end_time - start_time
            duration = int(elapsed)
            self.logger.info('Tempest Configuration successfully generated '
                             'in %s second at %s' % (duration, config_path))
        else:
            try:
                import config_tempest # noqa
                self.logging.warning('There is an error in syntax, please '
                                     'check $ refstack-client config -h')
            except ImportError:
                self.logger.warning('Please make sure python-tempestconf'
                                    'python package is installed')

    def test(self):
        '''Execute Tempest test against the cloud.'''
        self._prep_test()
        results_file = self._get_next_stream_subunit_output_file(
            self.tempest_dir)
        keystone_config = self._get_keystone_config(self.conf)
        auth_version, auth_url, content = \
            self._generate_keystone_data(keystone_config)
        cpid = self._get_cpid_from_keystone(auth_version, auth_url, content)

        self.logger.info("Starting Tempest test...")
        start_time = time.time()

        # Run tempest run command, conf file specified at _prep_test method
        # Use virtual environment (wrapper script)
        # Run the tests serially if parallel not enabled (--serial).
        wrapper = os.path.join(self.tempest_dir, 'tools', 'with_venv.sh')
        cmd = [wrapper, 'tempest', 'run']
        if not self.args.parallel:
            cmd.append('--serial')
        # If a test list was specified, have it take precedence.
        if self.args.test_list:
            self.logger.info("Normalizing test list...")
            parser = TestListParser(os.path.abspath(self.tempest_dir),
                                    insecure=self.args.insecure)
            # get include list
            list_file = parser.create_include_list(self.args.test_list)
            if list_file:
                if os.path.getsize(list_file) > 0:
                    # TODO(kopecmartin) rename the below argument when
                    # refstack-client uses tempest which contains the following
                    # change in its code:
                    # https://review.opendev.org/c/openstack/tempest/+/768583
                    cmd += ('--whitelist_file', list_file)
                else:
                    self.logger.error("Test list is either empty or no valid "
                                      "test cases for the tempest "
                                      "environment were found.")
                    exit(1)
            else:
                self.logger.error("Error normalizing passed in test list.")
                exit(1)
        elif 'arbitrary_args' in self.args:
            # Additional arguments for tempest run
            # otherwise run  all Tempest API tests.
            # keep usage(-- testCaseName)
            tmp = self.args.arbitrary_args[1:]
            if tmp:
                cmd += (tmp if tmp[0].startswith('-') else ['--regex'] + tmp)
        else:
            cmd += ['--regex', "tempest.api"]

        # If there were two verbose flags, show tempest results.
        if self.args.verbose > 0:
            stderr = None
        else:
            # Suppress tempest results output. Note that testr prints
            # results to stderr.
            stderr = open(os.devnull, 'w')

        # Execute the tempest run command in a subprocess.
        os.chdir(self.tempest_dir)
        process = subprocess.Popen(cmd, stderr=stderr)
        process.communicate()
        os.chdir(self.refstack_dir)

        # If the subunit file was created, then test cases were executed via
        # tempest run and there is test output to process.
        if os.path.isfile(results_file):
            end_time = time.time()
            elapsed = end_time - start_time
            duration = int(elapsed)

            self.logger.info('Tempest test complete.')
            self.logger.info('Subunit results located in: %s' % results_file)

            results = self.get_passed_tests(results_file)
            self.logger.info("Number of passed tests: %d" % len(results))

            content = self._form_result_content(cpid, duration, results)

            if self.args.result_tag:
                file_name = os.path.basename(results_file)
                directory = os.path.dirname(results_file)
                file_name = '-'.join([self.args.result_tag, file_name])
                results_file = os.path.join(directory, file_name)

            json_path = results_file + ".json"
            self._save_json_results(content, json_path)
            self.logger.info('JSON results saved in: %s' % json_path)

            # If the user specified the upload argument, then post
            # the results.
            if self.args.upload:
                self.post_results(self.args.url, content,
                                  sign_with=self.args.priv_key)
        else:
            msg1 = ("tempest run command did not generate a results file "
                    "under the Refstack os.path.dirname(results_file) "
                    "directory. Review command and try again.")
            msg2 = ("Problem executing tempest run command. Results file "
                    "not generated hence no file to upload. Review "
                    "arbitrary arguments.")
            if process.returncode != 0:
                self.logger.warning(msg1)
            if self.args.upload:
                self.logger.error(msg2)

        return process.returncode

    def upload(self):
        '''Perform upload to RefStack URL.'''
        self._prep_upload()
        json_file = open(self.upload_file)
        json_data = json.load(json_file)
        json_file.close()
        self._upload_prompt(json_data)

    def upload_subunit(self):
        '''Perform upload to RefStack URL from a subunit file.'''
        self._prep_upload()

        cpid = self._generate_cpid_from_endpoint(self.args.keystone_endpoint)
        # Forgo the duration for direct subunit uploads.
        duration = 0

        # Formulate JSON from subunit
        results = self.get_passed_tests(self.upload_file)
        self.logger.info('Number of passed tests in given subunit '
                         'file: %d ' % len(results))

        content = self._form_result_content(cpid, duration, results)
        self._upload_prompt(content)

    def yield_results(self, url, start_page=1,
                      start_date='', end_date='', cpid=''):
        endpoint = '%s/v1/results/' % url
        headers = {'Content-type': 'application/json'}
        for page in itertools.count(start_page):
            params = {'page': page}
            for param in ('start_date', 'end_date', 'cpid'):
                if locals()[param]:
                    params.update({param: locals()[param]})
            try:
                resp = requests.get(endpoint, headers=headers, params=params,
                                    verify=not self.args.insecure)
                resp.raise_for_status()
            except requests.exceptions.HTTPError as e:
                self.logger.info('Failed to list %s - %s ' % (endpoint, e))
                raise StopIteration
            else:
                resp = resp.json()
                results = resp.get('results', [])
                yield results
                if resp['pagination']['total_pages'] == page:
                    raise StopIteration

    def list(self):
        """Retrieve list with last test results from RefStack."""
        results = self.yield_results(self.args.url,
                                     start_date=self.args.start_date,
                                     end_date=self.args.end_date)
        for page_of_results in results:
            for r in page_of_results:
                print('%s - %s' % (r['created_at'], r['url']))
            try:
                moves.input('Press Enter to go to next page...')
            except KeyboardInterrupt:
                return

    def _sign_pubkey(self):
        """Generate self signature for public key"""
        private_key_file = self.args.priv_key_to_sign
        try:
            with open(private_key_file) as pkf:
                private_key = serialization.load_pem_private_key(
                    pkf.read().encode('utf-8'),
                    password=None,
                    backend=backends.default_backend())
        except (IOError, UnsupportedAlgorithm, ValueError) as e:
            self.logger.error('Error reading private key %s'
                              '' % private_key_file)
            self.logger.exception(e)
            return

        public_key_file = '.'.join((private_key_file, 'pub'))
        try:
            with open(public_key_file) as pkf:
                # Strip key comment at the end as it should not be included
                pub_key_elements = pkf.read().split(' ')
                pub_key = "%s %s" % (pub_key_elements[0], pub_key_elements[1])
        except IOError:
            self.logger.error('Public key file %s not found. '
                              'Public key is generated from private one.'
                              '' % public_key_file)
            pub_key = private_key.public_key().public_bytes(
                serialization.Encoding.OpenSSH,
                serialization.PublicFormat.OpenSSH)

        signature = private_key.sign(
            'signature'.encode('utf-8'),
            padding.PKCS1v15(),
            hashes.SHA256()
        )
        return pub_key, binascii.b2a_hex(signature)

    def self_sign(self):
        """Generate signature for public key."""
        pub_key, signature = self._sign_pubkey()
        print('Public key:\n%s\n' % pub_key)
        print('Self signature:\n%s\n' % str(signature, 'utf-8'))


def parse_cli_args(args=None):

    usage_string = ('refstack-client [-h] <ARG> ...\n\n'
                    'To see help on specific argument, do:\n'
                    'refstack-client <ARG> -h')

    parser = argparse.ArgumentParser(
        description='RefStack-client arguments',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        usage=usage_string
    )

    subparsers = parser.add_subparsers(help='Available subcommands.')

    # Arguments that go with all subcommands.
    shared_args = argparse.ArgumentParser(add_help=False)
    group = shared_args.add_mutually_exclusive_group()
    group.add_argument('-s', '--silent',
                       action='store_true',
                       help='Suppress output except warnings and errors.')

    group.add_argument('-v', '--verbose',
                       action='count',
                       default=0,
                       help='Show verbose output.')

    shared_args.add_argument('-y',
                             action='store_true',
                             dest='quiet',
                             required=False,
                             help='Assume Yes to all prompt queries')

    # Arguments that go with network-related  subcommands (test, list, etc.).
    network_args = argparse.ArgumentParser(add_help=False)
    network_args.add_argument('--url',
                              action='store',
                              required=False,
                              default=os.environ.get(
                                  'REFSTACK_URL',
                                  'https://refstack.openstack.org/api'),
                              type=str,
                              help='RefStack API URL to upload results to. '
                                   'Defaults to env[REFSTACK_URL] or '
                                   'https://refstack.openstack.org/'
                                   'api if it is not set '
                                   '(--url http://localhost:8000).')

    network_args.add_argument('-k', '--insecure',
                              action='store_true',
                              dest='insecure',
                              required=False,
                              help='Skip SSL checks while interacting '
                                   'with RefStack API and Keystone endpoint')

    network_args.add_argument('-i', '--sign',
                              type=str,
                              required=False,
                              dest='priv_key',
                              help='Path to private RSA key. '
                                   'OpenSSH RSA keys format supported')

    # Upload command
    parser_upload = subparsers.add_parser(
        'upload', parents=[shared_args, network_args],
        help='Upload an existing result JSON file.'
    )

    parser_upload.add_argument('file',
                               type=str,
                               help='Path of JSON results file.')

    parser_upload.set_defaults(func="upload")

    # Upload-subunit command
    parser_subunit_upload = subparsers.add_parser(
        'upload-subunit', parents=[shared_args, network_args],
        help='Upload results from a subunit file.'
    )

    parser_subunit_upload.add_argument('file',
                                       type=str,
                                       help='Path of subunit file.')

    parser_subunit_upload.add_argument('--keystone-endpoint',
                                       action='store',
                                       required=True,
                                       dest='keystone_endpoint',
                                       type=str,
                                       help='The Keystone URL of the cloud '
                                            'the subunit results belong to. '
                                            'This is used to generate a Cloud '
                                            'Provider ID.')

    parser_subunit_upload.set_defaults(func="upload_subunit")

    # Config Command
    parser_config = subparsers.add_parser(
        'config', parents=[shared_args, network_args],
        help='Generate tempest.conf for a cloud')

    parser_config.add_argument('--use-test-accounts',
                               action='store',
                               required=False,
                               dest='test_accounts',
                               type=str,
                               help='Path of the accounts.yaml file.')

    parser_config.add_argument('--network-id',
                               action='store',
                               required=False,
                               dest='network_id',
                               help='The ID of an existing network in our '
                                    'openstack instance with external '
                                    'connectivity')

    parser_config.add_argument('--image',
                               action='store',
                               required=False,
                               dest='image',
                               default=C.DEFAULT_IMAGE,
                               help='An image name chosen from `$ openstack '
                                    'image list` or a filepath/URL of an '
                                    'image to be uploaded to glance and set '
                                    'as a reference to be used by tests. The '
                                    'name of the image is the leaf name of '
                                    'the path. Default is %s'
                                    % C.DEFAULT_IMAGE)

    parser_config.add_argument('--out',
                               action='store',
                               required=False,
                               dest='out',
                               help='File path to write tempest.conf')

    parser_config.add_argument('--os-cloud',
                               action='store',
                               required=False,
                               dest='os_cloud',
                               help='Named cloud to connect to.')

    parser_config.add_argument('--overrides',
                               action='store',
                               required=False,
                               dest='overrides',
                               help='Comma seperated values which needs to be'
                                    'overridden in tempest.conf.'
                                    'Example --overrides'
                                    'compute.image_ref=<value>,'
                                    'compute.flavor_ref=<value>')

    parser_config.set_defaults(func='generate_tempest_config')

    # Test command
    parser_test = subparsers.add_parser(
        'test', parents=[shared_args, network_args],
        help='Run Tempest against a cloud.')

    parser_test.add_argument('-c', '--conf-file',
                             action='store',
                             required=True,
                             dest='conf_file',
                             type=str,
                             help='Path of the Tempest configuration file to '
                                  'use.')

    parser_test.add_argument('-r', '--result-file-tag',
                             action='store',
                             required=False,
                             dest='result_tag',
                             type=str,
                             help='Specify a string to prefix the result '
                                  'file with to easier distinguish them. ')

    parser_test.add_argument('--test-list',
                             action='store',
                             required=False,
                             dest='test_list',
                             type=str,
                             help='Specify the file path or URL of a test '
                                  'list text file. This test list will '
                                  'contain specific test cases that should '
                                  'be tested.')

    parser_test.add_argument('-u', '--upload',
                             action='store_true',
                             required=False,
                             help='After running Tempest, upload the test '
                                  'results to the default RefStack API server '
                                  'or the server specified by --url.')

    parser_test.add_argument('-p', '--parallel',
                             action='store_true',
                             required=False,
                             help='Run the tests in parallel. Note this '
                                  'requires multiple users/projects in '
                                  'accounts.yaml.')

    # This positional argument will allow arbitrary arguments to be passed in
    # with the usage of '--'.
    parser_test.add_argument('arbitrary_args',
                             nargs=argparse.REMAINDER,
                             help='After the first "--", you can pass '
                                  'arbitrary arguments to the tempest run '
                                  'runner. This can be used for running '
                                  'specific test cases or test lists. '
                                  'Some examples are: -- --regex '
                                  'tempest.api.compute.images.'
                                  'test_list_image_filters')
    parser_test.set_defaults(func="test")

    # List command
    parser_list = subparsers.add_parser(
        'list', parents=[shared_args, network_args],
        help='List last results from RefStack')
    parser_list.add_argument('--start-date',
                             required=False,
                             dest='start_date',
                             type=str,
                             help='Specify a date for start listing of '
                                  'test results '
                                  '(e.g. --start-date "2015-04-24 01:23:56").')
    parser_list.add_argument('--end-date',
                             required=False,
                             dest='end_date',
                             type=str,
                             help='Specify a date for end listing of '
                                  'test results '
                                  '(e.g. --end-date "2015-04-24 01:23:56").')
    parser_list.set_defaults(func='list')

    # Sign command
    parser_sign = subparsers.add_parser(
        'sign', parents=[shared_args],
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        help='Generate signature for public key.')
    parser_sign.add_argument('priv_key_to_sign',
                             type=str,
                             help='Path to private RSA key. '
                                  'OpenSSH RSA keys format supported')

    parser_sign.set_defaults(func='self_sign')

    return parser.parse_args(args=args)


def entry_point():
    args = parse_cli_args()
    test = RefstackClient(args)
    raise SystemExit(getattr(test, args.func)())



import os
import subprocess
import unittest


def get_project_path():
    path = os.path.dirname(os.path.realpath(__file__))
    while 'setup.py' not in os.listdir(path):
        path = os.path.realpath(os.path.join(path, '..'))
    return path


class TestSequenceFunctions(unittest.TestCase):
    scp_command = ('scp %s:/opt/stack/tempest/etc/'
                   'tempest.conf %s')
    pull_command = ('docker pull %s')
    test_command = ('docker run -t -v %s:/refstack-client -w /refstack-client/'
                    'refstack_client/tests/smoke --rm %s ./run_in_docker')

    def run_test(self, distro):
        subprocess.check_call(self.pull_command % distro, shell=True)
        subprocess.check_call(self.test_command % (get_project_path(), distro),
                              shell=True)

    def setUp(self):
        devstack_host = os.environ.get('DEVSTACK_HOST', None)
        self.assertIsNotNone(devstack_host)
        subprocess.check_call(
            self.scp_command % (devstack_host, get_project_path()),
            shell=True
        )

    def test_ubuntu_14(self):
        distro_image = 'ubuntu:12.04'
        self.run_test(distro_image)

    def test_ubuntu_12(self):
        distro_image = 'ubuntu:14.04'
        self.run_test(distro_image)

    def test_centos6(self):
        distro_image = 'centos:centos6'
        self.run_test(distro_image)

    def test_centos7(self):
        distro_image = 'centos:centos7'
        self.run_test(distro_image)

    def test_fedora_21(self):
        distro_image = 'fedora:21'
        self.run_test(distro_image)

    def test_opensuse_13(self):
        # offcial opensuse image has outdated certificates
        # we can't use it while this issue isn't fixed
        distro_image = 'opensuse/13.2'
        self.run_test(distro_image)


if __name__ == '__main__':
    unittest.main()
